// SPDX-License-Identifier: MIT

package main

import (
	"encoding/json"
	"encoding/xml"
	"fmt"
	"io/ioutil"
	"log"
	"net/http"
	"os"
	"sort"
	"strconv"
	"time"
)

type PkgDbCollections struct {
	Collections []Collection `json:"collections"`
	Output      string       `json:"output"`
}

type Collection struct {
	AllowRetire bool   `json:"allow_retire"`
	Branchname  string `json:"branchname"`
	DateCreated string `json:"date_created"`
	DateUpdated string `json:"date_updated"`
	DistTag     string `json:"dist_tag"`
	KojiName    string `json:"koji_name"`
	Name        string `json:"name"`
	Status      string `json:"status"`
	Version     string `json:"version"`
}

type Component struct {
	XMLName     xml.Name `xml:"component"`
	Type        string   `xml:"type,attr"`
	ID          string   `xml:"id"`
	Name        string   `xml:"name"`
	Summary     string   `xml:"summary"`
	Description struct {
		P string `xml:"p"`
	} `xml:"description"`
	URL struct {
		Text string `xml:",chardata"`
		Type string `xml:"type,attr"`
	} `xml:"url"`
	MetadataLicense string `xml:"metadata_license"`
	DeveloperName   string `xml:"developer_name"`
	Releases        struct {
		Release []Release `xml:"release"`
	} `xml:"releases"`
}

type Release struct {
	Version     string `xml:"version,attr"`
	Type        string `xml:"type,attr"`
	Date        string `xml:"date,attr"`
	DateEol     string `xml:"date_eol,attr"`
	Description struct {
		P string `xml:"p"`
	} `xml:"description"`
}

// Autogenerated from https://fedorapeople.org/groups/schedule/f-40/f-40-key.json
// using https://mholt.github.io/json-to-go/
type Schedule struct {
	Changelog   Changelog `json:"changelog"`
	End         string    `json:"end"`
	ExtAttr     ExtAttr   `json:"ext_attr"`
	FlagsAttrID any       `json:"flags_attr_id"`
	Name        string    `json:"name"`
	Resources   Resources `json:"resources"`
	Slug        string    `json:"slug"`
	Start       string    `json:"start"`
	Tasks       []Tasks   `json:"tasks"`
	UsedFlags   []string  `json:"used_flags"`
}
type Changelog struct {
}
type ExtAttr struct {
	Text1 int `json:"Text1"`
	Text2 int `json:"Text2"`
	Text3 int `json:"Text3"`
}
type Resources struct {
}
type Tasks struct {
	Level      int      `json:"_level"`
	Complete   float64  `json:"complete"`
	End        string   `json:"end"`
	Flags      []string `json:"flags"`
	Index      int      `json:"index"`
	Name       string   `json:"name"`
	ParentTask string   `json:"parentTask"`
	Priority   int      `json:"priority"`
	Slug       string   `json:"slug"`
	Start      string   `json:"start"`
	Tasks      []Tasks  `json:"tasks"`
	Type       string   `json:"type"`
}

type Bodhi struct {
	Name                   string `json:"name"`
	LongName               string `json:"long_name"`
	Version                string `json:"version"`
	IDPrefix               string `json:"id_prefix"`
	Branch                 string `json:"branch"`
	DistTag                string `json:"dist_tag"`
	StableTag              string `json:"stable_tag"`
	TestingTag             string `json:"testing_tag"`
	CandidateTag           string `json:"candidate_tag"`
	PendingSigningTag      string `json:"pending_signing_tag"`
	PendingTestingTag      string `json:"pending_testing_tag"`
	PendingStableTag       string `json:"pending_stable_tag"`
	OverrideTag            string `json:"override_tag"`
	MailTemplate           string `json:"mail_template"`
	State                  string `json:"state"`
	ComposedByBodhi        bool   `json:"composed_by_bodhi"`
	CreateAutomaticUpdates bool   `json:"create_automatic_updates"`
	PackageManager         string `json:"package_manager"`
	TestingRepository      string `json:"testing_repository"`
	Eol                    string `json:"eol"`
}

// Appstream metadata filename
var metadataFileName string = "org.fedoraproject.fedora.metainfo.xml"

// pkgdb collections JSON URL to get the list of releases
var pkgdb string = "https://admin.fedoraproject.org/pkgdb/api/collections/"

// URL template to get the release schedule for a given Fedora Linux release
var scheduleURL string = "https://fedorapeople.org/groups/schedule/f-%s/f-%s-key.json"

// URL template to get the EOL date for a given Fedora Linux release
var bodhiURL string = "https://bodhi.fedoraproject.org/releases/F%s"

// Oldest non-EOLed release to update
// Bump this value once it goes EOL
var oldestRelease = 39

func getJSON(url string) []byte {
	client := http.Client{}
	req, err := http.NewRequest(http.MethodGet, url, nil)
	if err != nil {
		log.Fatal(err)
	}
	res, err := client.Do(req)
	if err != nil {
		log.Fatal(err)
	}
	if res.Body != nil {
		defer res.Body.Close()
	}
	body, err := ioutil.ReadAll(res.Body)
	if err != nil {
		log.Fatal(err)
	}
	return body
}

func isValidURL(url string) bool {
	client := http.Client{}
	req, err := http.NewRequest(http.MethodGet, url, nil)
	if err != nil {
		return false
	}
	res, err := client.Do(req)
	if err != nil {
		return false
	}
	return res.StatusCode == 200
}

func getStartDateRecursive(tasks []Tasks) string {
	for _, t := range tasks {
		if t.Name == "Current Final Target date" && t.Type == "Milestone" {
			res, err := strconv.Atoi(t.Start)
			if err != nil {
				log.Fatal(err)
			}
			tm := time.Unix(int64(res), 0)
			// log.Printf("Current Final Target date: %d-%02d-%02d", tm.Year(), tm.Month(), tm.Day())
			return fmt.Sprintf("%d-%02d-%02d", tm.Year(), tm.Month(), tm.Day())
		} else {
			if len(t.Tasks) > 0 {
				date := getStartDateRecursive(t.Tasks)
				if date != "" {
					return date
				}
			}
		}
	}
	return ""
}

// Get the release date from the schedule
func getStartDate(version string) string {
	schedRaw := getJSON(fmt.Sprintf(scheduleURL, version, version))
	schedule := Schedule{}
	err := json.Unmarshal(schedRaw, &schedule)
	if err != nil {
		log.Fatal(err)
	}
	return getStartDateRecursive(schedule.Tasks)
}

// Get the End fo Life date from Bodhi
func getEolDate(version string) string {
	body := getJSON(fmt.Sprintf(bodhiURL, version))
	var bodhi Bodhi
	err := json.Unmarshal([]byte(body), &bodhi)
	if err != nil {
		log.Fatal(err)
	}
	return bodhi.Eol
}

func main() {
	// Read pkgdb collections from file if provided, otherwise fetch it
	var body []byte
	if len(os.Args[1:]) > 0 {
		jsonFile, err := os.Open(os.Args[1])
		if err != nil {
			fmt.Println(err)
		}
		defer jsonFile.Close()
		body, _ = ioutil.ReadAll(jsonFile)
	} else {
		body = getJSON(pkgdb)
	}
	collections := PkgDbCollections{}
	err := json.Unmarshal([]byte(body), &collections)
	if err != nil {
		log.Fatal(err)
	}
	if collections.Output != "ok" {
		log.Fatalln("Invalid value for output in JSON document")
	}

	// Read and parse existing metadata to keep historicall release dates and
	// only update the latest releases.
	f, err := os.Open(metadataFileName)
	if err != nil {
		log.Fatal(err)
	}
	body, _ = ioutil.ReadAll(f)
	f.Close()
	var m Component
	xml.Unmarshal(body, &m)

	// Basic sanity checks
	if m.Type != "operating-system" ||
		m.ID != "org.fedoraproject.fedora" ||
		m.Name != "Fedora Linux" ||
		m.URL.Text != "https://fedoraproject.org/" ||
		m.URL.Type != "homepage" ||
		m.MetadataLicense != "MIT" ||
		m.DeveloperName != "The Fedora Project" {
		log.Fatalln("Invalid metadata file. Reset to a valid file")
	}

	// Store all existing releases
	releases := make(map[string]Release)
	for _, r := range m.Releases.Release {
		releases[r.Version] = r
	}
	// Keep only Rawhide first as there are no dates to update
	m.Releases.Release = []Release{}
	m.Releases.Release = append(m.Releases.Release, releases["rawhide"])
	delete(releases, "rawhide")

	// Iterate through latest releases to update their start & eol dates
	for _, c := range collections.Collections {
		// Filter only Fedora Linux to ignore duplicates following the rename
		if c.Name != "Fedora Linux" {
			continue
		}
		// Ignore Rawhide
		if c.KojiName == "Rawhide" || c.Version == "devel" {
			continue
		}
		version, err := strconv.Atoi(c.Version)
		if err != nil {
			log.Fatal(err)
		}
		// Ignore too old End of Life releases no longer available in Bodhi
		if version < oldestRelease {
			continue
		}

		log.Printf("Updating dates for: %s %s", c.Name, c.Version)

		// Get info (if it exists) from the current metadata
		r := releases[c.Version]

		r.Version = c.Version
		r.Date = getStartDate(c.Version)
		r.DateEol = getEolDate(c.Version)

		if c.Status == "Under Development" {
			r.Type = "development"
			betaurl := fmt.Sprintf("https://fedoramagazine.org/announcing-fedora-%s-beta/", c.Version)
			if isValidURL(betaurl) {
				r.Description.P = betaurl
			} else {
				r.Description.P = "The upcoming release of Fedora Linux."
			}
		} else if c.Status == "Active" || c.Status == "EOL" {
			r.Type = "stable"
			if version < 40 {
				r.Description.P = fmt.Sprintf("https://fedoramagazine.org/announcing-fedora-%s/", c.Version)
			} else {
				r.Description.P = fmt.Sprintf("https://fedoramagazine.org/announcing-fedora-linux-%s/", c.Version)
			}
		} else {
			log.Fatal("Should never be reached")
		}

		log.Printf("Dates for %s %s (%s): Start: %s EOL: %s", c.Name, r.Version, r.Type, r.Date, r.DateEol)

		// Update map of releases
		releases[c.Version] = r
	}

	// Add back all releases to the XML metadata in reverse order
	var keys []string
	for k := range releases {
		keys = append(keys, k)
	}
	sort.Sort(sort.Reverse(sort.StringSlice(keys)))
	for _, k := range keys {
		m.Releases.Release = append(m.Releases.Release, releases[k])
	}

	// Update metadata file
	out, _ := xml.MarshalIndent(m, "", "  ")
	f, err = os.Create("org.fedoraproject.fedora.metainfo.xml")
	if err != nil {
		log.Fatal(err)
	}
	defer f.Close()
	_, err = f.WriteString("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n")
	if err != nil {
		log.Fatal(err)
	}
	_, err = f.Write(out)
	if err != nil {
		log.Fatal(err)
	}
	f.WriteString("\n")
	f.Sync()
}
